/*
 *	NEWS - GTK-based RSS Ticker - Copyright (C) Emmanuel Thomas-Maurin 2009-2011
 *	<manutm007@gmail.com>
 *
 * 	This program is free software: you can redistribute it and/or modify
 * 	it under the terms of the GNU General Public License as published by
 * 	the Free Software Foundation, either version 3 of the License, or
 * 	(at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * 	GNU General Public License for more details.
 *
 * 	You should have received a copy of the GNU General Public License
 * 	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "news.h"

#ifndef LIBXML_TREE_ENABLED
# error Libxml2: tree support not compiled in
#endif

#ifdef G_OS_WIN32
extern FILE	*stdout_fp, *stderr_fp;
#endif

static int	depth;
static xmlNode	*item_element;
static int	n;
/*
 * look for url and, if valid, parse it then dump result into
 * '<user_home_dir>/'NEWS_DIR_NAME/XMLDUMP'
 * if url not valid, only set error code and return
 */
int get_rss_feed(Resource *resrc, const Params *prm, int *rss_ttl)
{
	char	feed_title[FEED_TITLE_MAXLEN + 1];
	char	feed_link[FILENAME_MAXLEN + 1];
	char	feed_ttl[32];
	int	exit_status, i;

	get_gr_env()->suspend_rq = TRUE;
	*rss_ttl = prm->rss_refresh;

	str_n_cpy(resrc->xmldump, get_datafile_full_name_from_name(XMLDUMP), FILENAME_MAXLEN);

	if (resrc->fp != NULL)
		fclose(resrc->fp);
	resrc->fp = open_new_datafile_with_name(XMLDUMP, "wb");

	for (i = 0; i < NFEEDLINKANDOFFSETMAX; i++) {
		resrc->link_and_offset[i].offset_in_pixmap = 0;
		(resrc->link_and_offset[i].url)[0] = '\0';
	}

	if ((exit_status = parse_rss_xml_file(resrc->fp, resrc->id, resrc->link_and_offset, prm)) != OK)
		return exit_status;

	fclose(resrc->fp);
	resrc->fp = open_new_datafile_with_name(XMLDUMP, "rb");

	get_rss_feed_info(resrc->id, feed_title, feed_link, feed_ttl);
	str_n_cpy(resrc->feed_title, feed_title, FEED_TITLE_MAXLEN);
	if (feed_ttl[0] != '\0')
		*rss_ttl = atoi(feed_ttl);
	get_gr_env()->suspend_rq = FALSE;
	return OK;
}

/*
 * must be utf-8 encoded
 */
int parse_rss_xml_file(FILE *fp, const char* url, FeedLinkAndOffset *link_and_offset, const Params *prm)
{
	xmlDoc	*doc;
	xmlNode	*root_element;

	fprintf(STD_OUT, "Parsing XML file... ");
	if ((doc = xmlParseFile(url)) == NULL) {
		warning("XML parser error:", xmlGetLastError()->message, "", "");
		return RSS_UNPARSABLE;
	}
	if ((root_element = xmlDocGetRootElement(doc)) == NULL) {
		xmlFreeDoc(doc);
		warning("Empty XML document:", url, "", "");
		return RSS_EMPTY;
	}

	depth = 0;
	item_element = NULL;
	n = 1;

	get_rss_selected_elements1(root_element, doc);
	if (item_element != NULL)
		get_rss_selected_elements2(item_element, doc, fp, link_and_offset, prm);
	else {
		xmlFreeDoc(doc);
		warning("No 'Item' element found in:", url,
			"\nIs feed format RSS 2 ?\n",
			"(RSS 1 / Atom formats are not currently supported)");
		return RSS_NO_ITEM_ELEMENT;
	}
	xmlFreeDoc(doc);
	fprintf(STD_OUT, "done\n");
	return OK;
}

/*
 * for every link found, we insert in text LINKTAG_CHAR"00n"
 * with n = link rank and we fill link_and_offset with url
 * this is used later in render_stream_to_hpixmap()
 */
void get_rss_selected_elements1(xmlNode *some_element, xmlDoc *doc)
{
	xmlNode	*cur_node;

	for (cur_node = some_element; cur_node != NULL; cur_node = cur_node->next) {
		if (item_element != NULL)
			return;

		if (xmlStrcmp(cur_node->name, (const xmlChar *)"rss") == 0 && depth == 0)
			depth = 1;
		else if (xmlStrcmp(cur_node->name, (const xmlChar *)"channel") == 0 && depth == 1)
			depth = 2;
		else if (xmlStrcmp(cur_node->name, (const xmlChar *)"item") == 0 && depth == 2)
			depth = 3;

		if (depth == 3) {
			item_element = cur_node;
			return;
		} else
			get_rss_selected_elements1(cur_node->children, doc);
	}
	return;
}

void get_rss_selected_elements2(xmlNode *some_element, xmlDoc *doc,
		FILE *fp, FeedLinkAndOffset *link_and_offset, const Params *prm)
{
	xmlNode	*cur_node, *cur_node_bak;
	xmlChar	*str;

	for (cur_node = some_element; cur_node != NULL; cur_node = cur_node->next) {
		if (xmlStrcmp(cur_node->name, (const xmlChar *)"item") == 0) {
			cur_node_bak = cur_node;
			cur_node = cur_node->children;
			for (; cur_node != NULL; cur_node = cur_node->next) {
				if (prm->rss_title == 'y' && xmlStrcmp(cur_node->name, (const xmlChar *)"title") == 0) {
					if ((str = xmlNodeListGetString(doc, cur_node->children, 1)) != NULL) {
						/* we remove any LINKTAG_CHAR from str because it will be used in "link tag" */
						remove_char_from_str((char *)str, LINKTAG_CHAR);
						fprintf(fp, "%s%s\n", str, prm->rss_title_delimiter);
						xmlFree(str);
					}
				} else if (prm->rss_description == 'y' && xmlStrcmp(cur_node->name, (const xmlChar *)"description") == 0) {
					if ((str = xmlNodeListGetString(doc, cur_node->children, 1)) != NULL) {
						/* we remove any LINKTAG_CHAR from str because it will be used in "link tag" */
						remove_char_from_str((char *)str, LINKTAG_CHAR);
						fprintf(fp, "%s%s\n", str, prm->rss_description_delimiter);
						xmlFree(str);
					}
				}
			}
			cur_node = cur_node_bak;
			cur_node = cur_node->children;
			for (; cur_node != NULL; cur_node = cur_node->next) {
				if (xmlStrcmp(cur_node->name, (const xmlChar *)"link") == 0) {
					if ((str = xmlNodeListGetString(doc, cur_node->children, 1)) != NULL) {
						if (n < NFEEDLINKANDOFFSETMAX + 1) {
							str_n_cpy((link_and_offset + n)->url, (const char *)str, FILENAME_MAXLEN);
							fprintf(fp, "%c%03d\n", LINKTAG_CHAR, n++);
						}
						xmlFree(str);
					}
				}
			}
			cur_node = cur_node_bak;
		}
	}
	return;
}

/*
 * info is 4 strings 255 chars long each
 * feed_* can be NULL
 */
int get_rss_feed_info(const char *url, char *feed_title, char *feed_link, char *feed_ttl)
{
	xmlDoc	*doc;
	xmlNode	*root_element;

	if (feed_title != NULL)
		feed_title[0] = '\0';
	if (feed_link != NULL)
		feed_link[0] = '\0';
	if (feed_ttl != NULL)
		feed_ttl[0] = '\0';

	if ((doc = xmlParseFile(url)) == NULL) {
		return RSS_UNPARSABLE;
	} else {
		if ((root_element = xmlDocGetRootElement(doc)) == NULL) {
			xmlFreeDoc(doc);
			return RSS_EMPTY;
		} else {
			if (feed_title != NULL)
				get_xml_first_element(root_element->children, doc, "title", feed_title, FEED_TITLE_MAXLEN);
			if (feed_link != NULL)
				get_xml_first_element(root_element->children, doc, "link", feed_link, FILENAME_MAXLEN);
			if (feed_ttl != NULL)
				get_xml_first_element(root_element->children, doc, "ttl", feed_ttl, 31);
			xmlFreeDoc(doc);
			return OK;
		}
	}
}

void get_xml_first_element(xmlNode *some_element, xmlDoc *doc, char *name, char *string, int length)
{
	xmlNode	*cur_node;
	xmlChar	*str;

	for (cur_node = some_element; cur_node != NULL; cur_node = cur_node->next) {
		if (xmlStrcmp(cur_node->name, (const xmlChar *)name) == 0) {
			if ((str = xmlNodeListGetString(doc, cur_node->children, 1)) != NULL) {
				str_n_cpy(string, (const char *)str, length);
				xmlFree(str);
			} else
				string[0] = '\0';
			break;
		}
		get_xml_first_element(cur_node->children, doc, name, string, length);
	}
	return;
}
